from _typeshed import Incomplete

KWIMAGE_DISABLE_WARNINGS: Incomplete
KWIMAGE_DISABLE_TRANSFORM_WARNINGS: Incomplete
KWIMAGE_DISABLE_IMPORT_WARNINGS: Incomplete
KWIMAGE_DISABLE_TORCHVISION_NMS: Incomplete
KWIMAGE_DISABLE_C_EXTENSIONS: Incomplete
