import scriptconfig as scfg
from _typeshed import Incomplete


class StackImagesConfig(scfg.DataConfig):
    input_fpaths: Incomplete
    axis: Incomplete
    pad: Incomplete
    out: Incomplete


def main(cmdline: int = ..., **kwargs) -> None:
    ...
